#if 0
if uname -n |grep sfwr >/dev/null
then
cc  -DBRAIN_DEAD_SCO $0 -lsocket -o `basename $0 .c`
else
cc $0 -lsocket -lnsl -o `basename $0 .c`
fi
exit
#endif
/*
 *	rdate - fetch the time of day from a remote host and set ours
 *
 *	Usage: rdate [-a | -s] [-t tolerance] host ...
 *
 *	Rdate requests the current time from UDP port 37 on the given
 *	hosts until it succeeds.  It makes a crude attempt to correct
 *	for network propagation delay.  It then displays the time
 *	reported by the remote host plus the number of seconds difference
 *	between the remote host and the local host.  A positive 
 *	difference means the remote host is head of the local host.
 *
 *	If the -a or -s flags are set and the caller is the superuser,
 *	and if the difference between remote time and local time is
 *	more than a specified tolerance (default is 3 sec), then the
 *	local host's time is adjusted (-a) or set (-s) to match.
 *
 *	This program should not be expected to be accurate to less
 *	than a few seconds.
 *
 *	(c) Copyright 1989 by the Trustees of Dartmouth College.
 *	This program may be copied and used for any non-commercial
 *	purpose so long as this notice remains attached.
 *
 *	$Header: rdate.c,v 1.2 89/03/23 17:16:21 steve Exp $
 */
#include <sys/types.h>
#include <sys/socket.h>
#ifndef BEOS
#include <netinet/in.h>
#include <arpa/inet.h>
#endif
#include <netdb.h>
/* #include <strings.h> */
#include <sys/time.h>
#include <stdio.h>

#define NOADJTIME	/* if your system doesnt have adjtime(2) */
/* #define OLDSELECT	/* if you have the pre-fd_set select(2) */
/* #define OLDGETHOST	/* if your system has the 1-host_addr gethostbyname */

/* bcopy and friends are old-style - use memcpy and friends instead */
#define bcopy(s1,s2,len) memcpy(s2,s1,len)

#define DEFTOL		3		/* min delta we care about */
#define READTOUT	(long)10	/* timeout for packet read */

#ifdef OLDSELECT
#define FD_ZERO(sp)	((sp)->fds_bits[0] = 0)
#define FD_SET(fd,sp)	((sp)->fds_bits[0] |= (1<<(fd)))
#endif

int sflag = 0;
int aflag = 0;
int sock;
struct sockaddr_in server;
int tolerance = DEFTOL;

main(argc,argv)
	int argc;
	char *argv[];
{
	struct hostent *hp;
	struct servent *sp;
	char **ap;
	int errflag = 0;
	int c;
	extern char *optarg;
	extern int optind;

	while ((c = getopt(argc,argv,"ast:")) != EOF) {
		switch (c) {
		case 'a':
			if (sflag)
				errflag++;
			aflag++;
			break;
		case 's':
			if (aflag)
				errflag++;
			sflag++;
			break;
		case 't':
			tolerance = atoi(optarg);
			if (tolerance < 0)
				errflag++;
			break;
		default:
			errflag++;
		}
	}
	if (errflag || optind == argc) {
		fputs("Usage: rdate [-s | -a] [-t tolerance] host ...\n", stderr);
		exit(1);
	}

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("rdate: socket");
		exit(1);
	}
	server.sin_family = AF_INET;
#ifdef BEOS
	// Cannot find a services file, so have to hard code.
	server.sin_port = 37;
#else
	sp = getservbyname("time","udp");
	if (sp == NULL) {
		fprintf(stderr,"rdate: service time/udp is unknown.\n");
		exit(1);
	}
	server.sin_port = sp->s_port;
#endif
	for (; optind < argc; optind++) {
		/*
		 * See if an explicit addr was given. If not, try a hostname.
		 */
		server.sin_addr.s_addr = inet_addr(argv[optind]);
		if (server.sin_addr.s_addr != (unsigned long)-1) {
			fprintf(stdout,"rdate: Trying %s... ", argv[optind]);
			(void)fflush(stdout);
			if (doit())
				exit(0);
		} else {
			hp = gethostbyname(argv[optind]);
			if (hp == NULL) {
				fprintf(stderr,"rdate: host %s is unknown.\n", argv[optind]);
				continue;
			}
			/*
			 * There are 2 versions of the struct returned by
			 * gethostbyname: one has a single host address; the
			 * other has a list.  
			 */
#ifdef OLDGETHOST
			ap = &(hp->h_addr); {
#else
			for (ap = hp->h_addr_list; *ap; ap++) {
#endif
				bcopy((char *)*ap, (char *)&server.sin_addr, hp->h_length);
				fprintf(stdout,"rdate: Trying %s (%s)... ",
					hp->h_name, inet_ntoa(server.sin_addr));
				(void)fflush(stdout);
				if (doit())
					exit(0);
			}
		}
	}
	exit(1);	/* all failed */
}


int
doit()
{
	int retval = 1;
	fd_set readfds;
	int selval;
	int len;
	int delta;
	u_long rtime;
	char asciitime[26];
	struct sockaddr tmp_sockaddr;
	struct timeval tv, btv, ntv, stv;
#ifndef ntohl
	u_long ntohl();      /* declaration unneeded if ntohl is null macro */
#endif

	if (connect(sock, (struct sockaddr *) &server, sizeof(server)) < 0 ) {
		perror("\nrdate: connect");
		return(0);
	}
	/*
	 * Time the connection to estimate roundtrip time.
	 */
	if (gettimeofday(&btv,(struct timezone *)0) < 0) {
		perror("\nrdate: gettimeofday");
		return(0);
	}
	/*
	 * The protocol [RFC 868] calls for a null packet
	 * to be sent to the server first. Then it sends the
	 * time of day value.
	 *
	 * Unless it's brain-dead SCO UNIX, which doesn't
	 * seem willing to send null-length packets.  Luckily,
	 * it seems the time servers are not so picky that
	 * they won't allow at least *one* data byte.
	 */
#if defined( BRAIN_DEAD_SCO )
	if (sendto(sock, (char *)&rtime, 1, 0, &tmp_sockaddr, sizeof( struct sockaddr )) < 0) {
#else
	if (send(sock, (char *)&rtime, sizeof( u_long ), 0) < 0) {
#endif
		perror("\nrdate: sendto");
		return(0);
	}
#ifndef BEOS
	FD_ZERO(&readfds);
	FD_SET(sock, &readfds);
	stv.tv_sec = READTOUT;
	stv.tv_usec = (long)0;
	if ((selval = select(sock+1, &readfds, (fd_set *)0, (fd_set *)0, &stv)) == 0) {
		fputs("read timed out.\n", stdout);
		return(0);
	} else if (selval < 0) {
		perror("\nrdate: select");
		return(0);
	}
	if (gettimeofday(&tv,(struct timezone *)0) < 0) {
		perror("\nrdate: gettimeofday");
		return(0);
	}
	if (read(sock, (char *)&rtime, sizeof(rtime)) < sizeof(rtime)) {
		perror("\nrdate: read");
		return(0);
	}
#else
	(void)alarm( READTOUT );	// Allow 2 seconds for command.
	if ( recv( sock, &rtime, sizeof( rtime ), 0) < sizeof( rtime ) ) {
		perror( "\nrdate: read" );
		return( 0);
	}
	if (gettimeofday(&tv,(struct timezone *)0) < 0) {
		perror("\nrdate: gettimeofday");
		return(0);
	}
#endif
	/*
	 * The inet time is seconds since 1 Jan 1900 and Unix's
	 * is seconds since 1 Jan 1970, so adjust.
	 */
	rtime = ntohl(rtime) - (u_long)2208988800;

	/*
	 * Adjust again by estimating propagation time from
	 * the remote host to be half the round trip time.
	 */
#ifdef DEBUG
	printf("\nbtv = %d + %d\n tv = %d + %d\nrtm = %d\n",
	   btv.tv_sec,btv.tv_usec,tv.tv_sec,tv.tv_usec,rtime);
#endif
	rtime += (1 + tv.tv_sec - btv.tv_sec ) / 2;

	delta = rtime - tv.tv_sec;
	if ((sflag || aflag) && abs(delta) > tolerance) {
		ntv.tv_usec = (long)0;
		if (sflag) {
			ntv.tv_sec = rtime;
#ifdef BEOS
			set_real_time_clock( rtime );
#else
			if (settimeofday(&ntv,(struct timezone *)0) < 0) {
				perror("\nrdate: settimeofday");
				retval = 0;
			}
#endif
		} else if (aflag) {
#ifdef NOADJTIME
			fputs("\nrdate: -a option not available.\n", stderr);
			retval = 0;
#else
			ntv.tv_sec = delta;
			if (adjtime(&ntv,(struct timeval *)0) < 0) {
				perror("\nrdate: adjtime");
				retval = 0;
			}
#endif
		}
	}

	(void)strcpy(asciitime,ctime((long *)&rtime));
	asciitime[24] = '\0';
	fprintf(stdout,"%s (%s%d)\n", asciitime, 
		delta > 0 ? "+" : "", delta);
	return(retval);
}
